"use strict";
var __defProp = Object.defineProperty;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __export = (target, all) => {
  for (var name in all)
    __defProp(target, name, { get: all[name], enumerable: true });
};
var __copyProps = (to, from, except, desc) => {
  if (from && typeof from === "object" || typeof from === "function") {
    for (let key of __getOwnPropNames(from))
      if (!__hasOwnProp.call(to, key) && key !== except)
        __defProp(to, key, { get: () => from[key], enumerable: !(desc = __getOwnPropDesc(from, key)) || desc.enumerable });
  }
  return to;
};
var __toCommonJS = (mod) => __copyProps(__defProp({}, "__esModule", { value: true }), mod);
var __publicField = (obj, key, value) => {
  __defNormalProp(obj, typeof key !== "symbol" ? key + "" : key, value);
  return value;
};

// index.ts
var rushlight_exports = {};
__export(rushlight_exports, {
  CollabClient: () => CollabClient
});
module.exports = __toCommonJS(rushlight_exports);

// collab.ts
var import_collab2 = require("@codemirror/collab");
var import_state2 = require("@codemirror/state");
var import_view2 = require("@codemirror/view");

// presence.ts
var import_collab = require("@codemirror/collab");
var import_state = require("@codemirror/state");
var import_view = require("@codemirror/view");
function presenceToJSON({ selection, clientID }) {
  return {
    selection: selection.toJSON(),
    clientID
  };
}
function presenceFromJSON(value) {
  return {
    selection: import_state.EditorSelection.fromJSON(value.selection),
    clientID: value.clientID
  };
}
var addPresence = import_state.StateEffect.define({
  map({ selection, clientID }, changes) {
    return {
      selection: selection.map(changes),
      clientID
    };
  }
});
var CursorWidget = class extends import_view.WidgetType {
  constructor(hue) {
    super();
    this.hue = hue;
  }
  toDOM() {
    let el = document.createElement("span");
    el.setAttribute("aria-hidden", "true");
    el.style.position = "absolute";
    el.style.borderLeft = el.style.borderRight = `1px solid hsl(${this.hue} 90% 40%)`;
    el.style.marginTop = "1.5px";
    el.style.height = "1.2em";
    return el;
  }
  eq(widget) {
    return widget.hue === this.hue;
  }
  get estimatedHeight() {
    return 0;
  }
  ignoreEvent() {
    return false;
  }
};
function defaultDecorations(presences) {
  const decorations = [];
  for (const { selection, clientID } of presences) {
    let hue = 0;
    for (let i = 0; i < clientID.length; i++)
      hue = Math.imul(hue + clientID.charCodeAt(i), 595438061) % 360;
    const cursorWidget = import_view.Decoration.widget({
      widget: new CursorWidget(hue)
    });
    const underlineMark = import_view.Decoration.mark({
      attributes: {
        style: `background-color: hsl(${hue} 90% 40% / 0.15)`
      }
    });
    decorations.push(cursorWidget.range(selection.main.head));
    for (const range of selection.ranges) {
      if (!range.empty) {
        decorations.push(underlineMark.range(range.from, range.to));
      }
    }
  }
  return import_view.Decoration.set(decorations, true);
}
function presenceExtension(presenceInterval, presenceExpiry, presenceDecorations) {
  const presenceField = import_state.StateField.define({
    create() {
      return [];
    },
    update(presenceSet, tr) {
      const now = Date.now();
      const me = (0, import_collab.getClientID)(tr.state);
      const byId = /* @__PURE__ */ new Map();
      for (const e of tr.effects) {
        if (e.is(addPresence) && e.value.clientID !== me) {
          byId.set(e.value.clientID, { ...e.value, time: now });
        }
      }
      const newPresenceSet = [...byId.values()];
      for (const p of presenceSet) {
        if (p.time >= now - presenceExpiry && !byId.has(p.clientID)) {
          newPresenceSet.push({ ...p, selection: p.selection.map(tr.changes) });
        }
      }
      return newPresenceSet;
    },
    provide(self) {
      return import_view.EditorView.decorations.from(self, presenceDecorations);
    }
  });
  const presenceFilter = import_state.EditorState.transactionFilter.of((tr) => {
    if (tr.selection) {
      const effect = addPresence.of({
        selection: tr.selection,
        clientID: (0, import_collab.getClientID)(tr.state)
      });
      return { ...tr, effects: [...tr.effects, effect] };
    }
    return tr;
  });
  const presenceTimer = import_view.ViewPlugin.define((view) => {
    const sendPresence = () => {
      const effect = addPresence.of({
        selection: view.state.selection,
        clientID: (0, import_collab.getClientID)(view.state)
      });
      view.dispatch({ effects: effect });
    };
    const timer = setInterval(sendPresence, presenceInterval);
    return { destroy: () => clearInterval(timer) };
  });
  return [presenceField, presenceFilter, presenceTimer];
}

// collab.ts
function pushUpdates(connection, version, fullUpdates) {
  let updates = fullUpdates.map((u) => ({
    clientID: u.clientID,
    changes: u.changes.toJSON(),
    effects: u.effects?.map((e) => presenceToJSON(e.value))
  }));
  return connection({ type: "pushUpdates", version, updates });
}
async function pullUpdates(connection, version) {
  const resp = await connection({ type: "pullUpdates", version });
  if (resp.status === "desync") {
    return "desync";
  }
  return resp.updates.map((u) => ({
    changes: import_state2.ChangeSet.fromJSON(u.changes),
    clientID: u.clientID,
    effects: u.effects?.map((e) => addPresence.of(presenceFromJSON(e)))
  }));
}
async function getDocument(connection) {
  const data = await connection({ type: "getDocument" });
  return {
    version: data.version,
    doc: import_state2.Text.of(data.doc.split("\n"))
  };
}
function peerExtension(startVersion, connection) {
  let plugin = import_view2.ViewPlugin.fromClass(
    class {
      constructor(view) {
        this.view = view;
        __publicField(this, "pushing", false);
        __publicField(this, "done", false);
        __publicField(this, "failures", 0);
        this.pull();
      }
      update(update) {
        if (update.docChanged || update.transactions.some((tr) => tr.effects.length)) {
          setTimeout(() => this.push(), 0);
        }
      }
      // Increment the failure count, and sleep with exponential backoff if 3
      // requests fail in a row.
      async failureSleep() {
        this.failures++;
        if (this.failures >= 3) {
          const delay = Math.pow(1.5, Math.min(this.failures - 3, 7)) * 1e3;
          await new Promise((resolve) => setTimeout(resolve, delay));
        }
      }
      async push() {
        let updates = (0, import_collab2.sendableUpdates)(this.view.state);
        if (this.pushing || !updates.length)
          return;
        this.pushing = true;
        let version = (0, import_collab2.getSyncedVersion)(this.view.state);
        try {
          await pushUpdates(connection, version, updates);
          this.failures = 0;
        } catch (e) {
          console.error(e);
          await this.failureSleep();
        }
        this.pushing = false;
        if ((0, import_collab2.sendableUpdates)(this.view.state).length)
          setTimeout(() => this.push(), 100);
      }
      async pull() {
        while (!this.done) {
          const version = (0, import_collab2.getSyncedVersion)(this.view.state);
          try {
            const updates = await pullUpdates(connection, version);
            this.failures = 0;
            if (updates === "desync") {
              await this.fullReload();
            } else {
              this.view.dispatch((0, import_collab2.receiveUpdates)(this.view.state, updates));
            }
          } catch (e) {
            console.error(e);
            await this.failureSleep();
          }
        }
      }
      async fullReload() {
        const { version, doc } = await getDocument(connection);
        const oldVersion = (0, import_collab2.getSyncedVersion)(this.view.state);
        const oldLength = (0, import_collab2.sendableUpdates)(this.view.state).reduce(
          (sum, u) => sum - u.changes.newLength + u.changes.length,
          this.view.state.doc.length
        );
        const changeSets = [
          import_state2.ChangeSet.of([{ from: 0, to: oldLength, insert: doc }], oldLength)
        ];
        for (let i = 1; i < version - oldVersion; i++) {
          changeSets.push(import_state2.ChangeSet.empty(doc.length));
        }
        const updates = changeSets.map((c) => ({ changes: c, clientID: "" }));
        this.view.dispatch((0, import_collab2.receiveUpdates)(this.view.state, updates));
      }
      destroy() {
        this.done = true;
      }
    }
  );
  const collabExtension = (0, import_collab2.collab)({
    startVersion,
    sharedEffects: (tr) => tr.effects.filter((e) => e.is(addPresence))
  });
  return [collabExtension, plugin];
}

// index.ts
var defaultOptions = {
  presenceInterval: 5e3,
  presenceExpiry: 8e3,
  presenceDecorations: defaultDecorations
};
var CollabClient = class _CollabClient {
  constructor(initialDoc, extension, presence) {
    this.initialDoc = initialDoc;
    this.extension = extension;
    this.presence = presence;
  }
  static async of(options) {
    const opts = { ...defaultOptions, ...options };
    const { version, doc } = await getDocument(options.connection);
    return new _CollabClient(
      doc,
      peerExtension(version, options.connection),
      presenceExtension(
        opts.presenceInterval,
        opts.presenceExpiry,
        opts.presenceDecorations
      )
    );
  }
};
// Annotate the CommonJS export names for ESM import in node:
0 && (module.exports = {
  CollabClient
});
